//! Defines the structure of the `project.toml` file.

use crate::{
	iotro::Language,
	render::ffmpeg::FfmpegOutputFormat,
	time::{Date, Time}
};
use rational::Rational;
use serde::{Deserialize, Serialize};
use serde_with::{serde_as, DisplayFromStr};
use std::{
	cmp,
	collections::BTreeSet,
	fmt::{self, Display, Formatter},
	str::FromStr
};

#[derive(Clone, Copy, Debug, Deserialize, Serialize)]
pub struct Resolution(u32, u32);

impl Resolution {
	pub fn new(width: u32, height: u32) -> Self {
		Self(width, height)
	}

	pub fn width(self) -> u32 {
		self.0
	}

	pub fn height(self) -> u32 {
		self.1
	}

	pub(crate) fn bitrate(self) -> u64 {
		// 640 * 360: 500k
		if self.width() <= 640 {
			500_000
		}
		// 1280 * 720: 1M
		else if self.width() <= 1280 {
			1_000_000
		}
		// 1920 * 1080: 2M
		else if self.width() <= 1920 {
			2_000_000
		}
		// 2560 * 1440: 3M
		else if self.width() <= 2560 {
			3_000_000
		}
		// 3840 * 2160: 4M
		// TODO die bitrate von 4M ist absolut an den haaren herbeigezogen
		else if self.width() <= 3840 {
			4_000_000
		}
		// we'll cap everything else at 5M for no apparent reason
		else {
			5_000_000
		}
	}

	pub(crate) fn default_codec(self) -> FfmpegOutputFormat {
		if self.width() > 1920 {
			FfmpegOutputFormat::Av1Opus
		} else {
			FfmpegOutputFormat::AvcAac
		}
	}

	pub const STANDARD_RESOLUTIONS: [Self; 5] = [
		Self(640, 360),
		Self(1280, 720),
		Self(1920, 1080),
		Self(2560, 1440),
		Self(3840, 2160)
	];
}

impl Display for Resolution {
	fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
		write!(f, "{}p", self.height())
	}
}

impl FromStr for Resolution {
	type Err = anyhow::Error;

	fn from_str(s: &str) -> anyhow::Result<Self> {
		Ok(match s.to_lowercase().as_str() {
			"360p" | "nhd" => Self(640, 360),
			"540p" | "qhd" => Self(960, 540),
			"720p" | "hd" => Self(1280, 720),
			"900p" | "hd+" => Self(1600, 900),
			"1080p" | "fhd" | "fullhd" => Self(1920, 1080),
			"1440p" | "wqhd" => Self(2560, 1440),
			"2160p" | "4k" | "uhd" => Self(3840, 2160),
			lower => {
				let mut parts = lower.split('x');
				match (
					parts.next().map(|width| width.parse()),
					parts.next().map(|height| height.parse()),
					parts.next()
				) {
					(Some(Ok(width)), Some(Ok(height)), None) => Self(width, height),
					_ => anyhow::bail!("Unknown Resolution: {s:?}")
				}
			}
		})
	}
}

impl Ord for Resolution {
	fn cmp(&self, other: &Self) -> cmp::Ordering {
		(self.0 * self.1).cmp(&(other.0 * other.1))
	}
}

impl Eq for Resolution {}

impl PartialOrd for Resolution {
	fn partial_cmp(&self, other: &Self) -> Option<cmp::Ordering> {
		Some(self.cmp(other))
	}
}

impl PartialEq for Resolution {
	fn eq(&self, other: &Self) -> bool {
		self.cmp(other) == cmp::Ordering::Equal
	}
}

#[derive(Debug, Deserialize, Serialize)]
pub struct Project {
	pub lecture: ProjectLecture,
	pub source: ProjectSource,
	pub progress: ProjectProgress
}

#[serde_as]
#[derive(Debug, Deserialize, Serialize)]
pub struct ProjectLecture {
	pub course: String,
	pub label: String,
	pub docent: String,
	#[serde_as(as = "DisplayFromStr")]
	pub date: Date,
	#[serde(default = "Default::default")]
	#[serde_as(as = "DisplayFromStr")]
	pub lang: Language<'static>
}

#[serde_as]
#[derive(Debug, Deserialize, Serialize)]
pub struct ProjectSource {
	pub files: Vec<String>,
	pub stereo: bool,

	#[serde_as(as = "Option<DisplayFromStr>")]
	pub start: Option<Time>,
	#[serde_as(as = "Option<DisplayFromStr>")]
	pub end: Option<Time>,

	#[serde(default)]
	#[serde_as(as = "Vec<(DisplayFromStr, DisplayFromStr)>")]
	pub fast: Vec<(Time, Time)>,

	#[serde(default)]
	#[serde_as(as = "Vec<(DisplayFromStr, DisplayFromStr, _)>")]
	pub questions: Vec<(Time, Time, String)>,

	pub metadata: Option<ProjectSourceMetadata>
}

#[serde_as]
#[derive(Debug, Deserialize, Serialize)]
pub struct ProjectSourceMetadata {
	/// The duration of the source video.
	#[serde_as(as = "DisplayFromStr")]
	pub source_duration: Time,
	/// The FPS of the source video.
	#[serde_as(as = "DisplayFromStr")]
	pub source_fps: Rational,
	/// The time base of the source video.
	#[serde_as(as = "DisplayFromStr")]
	pub source_tbn: Rational,
	/// The resolution of the source video.
	pub source_res: Resolution,
	/// The sample rate of the source audio.
	pub source_sample_rate: u32
}

#[serde_as]
#[derive(Debug, Default, Deserialize, Serialize)]
pub struct ProjectProgress {
	#[serde(default)]
	pub preprocessed: bool,

	#[serde(default)]
	pub asked_start_end: bool,

	#[serde(default)]
	pub asked_fast: bool,

	#[serde(default)]
	pub asked_questions: bool,

	#[serde(default)]
	pub rendered_assets: bool,

	#[serde(default)]
	pub rendered: bool,

	#[serde_as(as = "BTreeSet<DisplayFromStr>")]
	#[serde(default)]
	pub transcoded: BTreeSet<Resolution>
}

#[cfg(test)]
mod tests {
	use super::*;

	#[test]
	fn parse_custom_resolution() {
		assert_eq!(
			Resolution(1920, 1334),
			"1920x1334".parse::<Resolution>().unwrap()
		);
	}
}
